@@ -1,4 +1,5 @@
 #
+# Copyright (C) 2006, 2007 Loic Dachary <loic@dachary.org>
 # Copyright (C) 2004, 2005, 2006 Mekensleep
 #
 # Mekensleep
@@ -26,7 +27,12 @@
 from math import ceil
 from types import StringType
 from pprint import pformat
-import time, sys
+import time, sys, random
+
+def tournament_seconds():
+    return time.time()
+
+shuffler = random
 
 from pokerengine.pokergame import PokerGameServer
 from pokerengine.pokerengineconfig import Config
@@ -34,7 +40,10 @@ from pokerengine.pokerengineconfig import Config
 TOURNAMENT_STATE_ANNOUNCED = "announced"
 TOURNAMENT_STATE_REGISTERING = "registering"
 TOURNAMENT_STATE_RUNNING = "running"
+TOURNAMENT_STATE_BREAK_WAIT = "breakwait"
+TOURNAMENT_STATE_BREAK = "break"
 TOURNAMENT_STATE_COMPLETE = "complete"
+TOURNAMENT_STATE_CANCELED = "canceled"
             
 def equalizeCandidates(games):
     #
@@ -170,6 +179,7 @@ class PokerTournament:
         self.serial = kwargs.get('serial', 1)
         self.verbose = kwargs.get('verbose', 0)
         self.players_quota = kwargs.get('players_quota', 10)
+        self.players_min = kwargs.get('players_min', 2)
         self.variant = kwargs.get('variant', 'holdem')
         self.betting_structure = kwargs.get('betting_structure', 'level-15-30-no-limit')
         self.dirs = kwargs.get('dirs', [])
@@ -177,11 +187,20 @@ class PokerTournament:
         self.sit_n_go = kwargs.get('sit_n_go', 'y')
         self.register_time = kwargs.get('register_time', 0)
         self.start_time = kwargs.get('start_time', 0)
-        self.breaks_interval = kwargs.get('breaks_interval', 60)
+        self.breaks_first = kwargs.get('breaks_first', 7200)
+        self.breaks_interval = kwargs.get('breaks_interval', 3600)
+        self.breaks_duration = kwargs.get('breaks_duration', 300)
+        self.breaks_running_since = -1
+        self.breaks_since = -1
+        self.breaks_count = 0
+        self.buy_in = int(kwargs.get('buy_in', 0))
+        self.rake = int(kwargs.get('rake', 0))
         self.rebuy_delay = kwargs.get('rebuy_delay', 0)
         self.add_on = kwargs.get('add_on', 0)
         self.add_on_delay = kwargs.get('add_on_delay', 60)
+        self.prize_min = kwargs.get('prize_min', 0)
         self.prizes_specs = kwargs.get('prizes_specs', "table")
+        self.rank2prize = None
         self.finish_time = -1
         if type(self.start_time) is StringType:
             self.start_time = int(time.mktime(time.strptime(self.start_time, "%Y/%m/%d %H:%M")))
@@ -196,12 +215,13 @@ class PokerTournament:
         self.games = []
         self.id2game = {}
         
-        self.callback_new_state = lambda tournament: True
+        self.callback_new_state = lambda tournament, old_state, new_state: True
         self.callback_create_game = lambda tournament: PokerGameServer("poker.%s.xml", tournament.dirs)
         self.callback_game_filled = lambda tournament, game: True
         self.callback_destroy_game = lambda tournament, game: True
         self.callback_move_player = lambda tournament, from_game_id, to_game_id, serial: self.movePlayer(from_game_id, to_game_id, serial)
         self.callback_remove_player = lambda tournament, game_id, serial: self.removePlayer(game_id, serial)
+        self.callback_cancel = lambda tournament: True
         self.loadPayouts()
         self.updateRegistering()
 
@@ -217,7 +237,18 @@ class PokerTournament:
         print self.prefix + "[PokerTournament %s] " % self.name + message
         
     def canRun(self):
-        return self.start_time < time.time() and self.sit_n_go == 'y' and self.players_quota <= self.registered
+        if self.start_time < tournament_seconds():
+            if self.sit_n_go == 'y' and self.registered >= self.players_quota:
+                return True
+            elif self.sit_n_go == 'n':
+                if self.registered >= self.players_min:
+                    return True
+                else:
+                    return None
+            else:
+                return False
+        else:
+            return False
 
     def getRank(self, serial):
         try:
@@ -229,7 +260,7 @@ class PokerTournament:
         
     def updateRegistering(self):
         if self.state == TOURNAMENT_STATE_ANNOUNCED:
-            now = time.time()
+            now = tournament_seconds()
             if now - self.register_time > 0.0:
                 self.changeState(TOURNAMENT_STATE_REGISTERING)
                 return -1
@@ -238,29 +269,92 @@ class PokerTournament:
         else:
             if self.verbose > 0: self.message("updateRegistering: should not be called while tournament is not in announced state")
             return -1
-            
+
+    def updateRunning(self):
+        if self.state == TOURNAMENT_STATE_REGISTERING:
+            ready = self.canRun()
+            if ready == True:
+                self.changeState(TOURNAMENT_STATE_RUNNING)
+            elif ready == None:
+                self.changeState(TOURNAMENT_STATE_CANCELED)
+            elif ready == False:
+                pass
+
+    def remainingBreakSeconds(self):
+        if self.breaks_since > 0:
+            return self.breaks_duration - ( tournament_seconds() - self.breaks_since )
+        else:
+            return None
+        
+    def updateBreak(self, game_id = None):
+        if self.breaks_duration <= 0:
+            return False
+
+        if self.state == TOURNAMENT_STATE_RUNNING:
+            running_duration = tournament_seconds() - self.breaks_running_since
+            if self.breaks_count > 0:
+                running_max = self.breaks_interval
+            else:
+                running_max = self.breaks_first
+            if running_duration >= running_max:
+                self.breaks_games_id = []
+                self.changeState(TOURNAMENT_STATE_BREAK_WAIT)
+                
+        if self.state == TOURNAMENT_STATE_BREAK_WAIT:
+            self.breaks_games_id.append(game_id)
+            on_break = True
+            for game in self.games:
+                if game.id not in self.breaks_games_id:
+                    on_break = False
+                    break
+            if on_break:
+                del self.breaks_games_id
+                self.changeState(TOURNAMENT_STATE_BREAK)
+
+        if self.state == TOURNAMENT_STATE_BREAK:
+            if self.remainingBreakSeconds() <= 0:
+                self.changeState(TOURNAMENT_STATE_RUNNING)
+
+        if self.state not in (TOURNAMENT_STATE_RUNNING, TOURNAMENT_STATE_BREAK_WAIT, TOURNAMENT_STATE_BREAK):
+            if self.verbose >= 0: print "PokerTournament:updateBreak: is not supposed to be called while in state %s" % self.state
+            return None
+        
+        return True
+        
     def changeState(self, state):
         if self.state == TOURNAMENT_STATE_ANNOUNCED and state == TOURNAMENT_STATE_REGISTERING:
             self.can_register = True
+        elif self.state == TOURNAMENT_STATE_RUNNING and state == TOURNAMENT_STATE_BREAK_WAIT:
+            pass
+        elif self.state == TOURNAMENT_STATE_BREAK_WAIT and state == TOURNAMENT_STATE_BREAK:
+            self.breaks_since = tournament_seconds()
+        elif self.state == TOURNAMENT_STATE_BREAK and state == TOURNAMENT_STATE_RUNNING:
+            self.breaks_since = -1
+            self.breaks_running_since = tournament_seconds()
         elif self.state == TOURNAMENT_STATE_REGISTERING and state == TOURNAMENT_STATE_RUNNING:
-            self.start_time = time.time()
+            self.start_time = tournament_seconds()
+            self.breaks_running_since = self.start_time
             self.createGames()
-            if self.sit_n_go == 'y' or self.registered >= self.players_quota:
-                self.can_register = False
-        elif self.state == TOURNAMENT_STATE_RUNNING and state == TOURNAMENT_STATE_COMPLETE:
-            self.finish_time = time.time()
+            self.can_register = False
+        elif self.state == TOURNAMENT_STATE_REGISTERING and state == TOURNAMENT_STATE_CANCELED:
+            self.cancel()
+            self.finish_time = tournament_seconds()
+        elif ( self.state in ( TOURNAMENT_STATE_RUNNING, TOURNAMENT_STATE_BREAK_WAIT ) and
+               state == TOURNAMENT_STATE_COMPLETE ):
+            self.finish_time = tournament_seconds()
         else:
             if self.verbose >= 0: print "PokerTournament:changeState: cannot change from state %s to state %s" % ( self.state, state )
             return
         if self.verbose > 2: self.message("state change %s => %s" % ( self.state, state ))
+        old_state = self.state
         self.state = state
-        self.callback_new_state(self)
+        self.callback_new_state(self, old_state, self.state)
 
     def isRegistered(self, serial):
         return serial in self.players
         
     def canRegister(self, serial):
-        if self.can_register:
+        if self.can_register and self.registered < self.players_quota:
             return not self.isRegistered(serial)
         else:
             return False
@@ -273,7 +367,7 @@ class PokerTournament:
             self.players.append(serial)
             self.registered += 1
             if self.state == TOURNAMENT_STATE_REGISTERING:
-                if self.canRun(): self.changeState(TOURNAMENT_STATE_RUNNING)
+                self.updateRunning()
             elif self.state == TOURNAMENT_STATE_RUNNING:
                 self.sitPlayer(serial)
             return True
@@ -288,6 +382,15 @@ class PokerTournament:
         else:
             return False
 
+    def cancel(self):
+        if self.state == TOURNAMENT_STATE_REGISTERING:
+            self.callback_cancel(self)
+            self.players = []
+            self.registered = 0
+            return True
+        else:
+            return False
+        
     def sitPlayer(self, serial):
         pass
 
@@ -318,6 +421,7 @@ class PokerTournament:
         games_count = int(ceil(self.registered / float(self.seats_per_game)))
         self.players_quota = games_count * self.seats_per_game
         players = self.players[:]
+        shuffler.shuffle(players)
         for id in xrange(1, games_count + 1):
             game = self.callback_create_game(self)
             game.verbose = self.verbose
@@ -370,6 +474,8 @@ class PokerTournament:
             return False
         else:
             if loosers_count > 0 or self.need_balance: self.balanceGames()
+            if self.id2game.has_key(game_id):
+                self.updateBreak(game_id)
             return True
         
     def balanceGames(self):
@@ -404,17 +510,18 @@ class PokerTournament:
         
         return len(to_equalize) > 0
 
-    def prizes(self, buy_in):
+    def prizes(self):
         if self.can_register:
             return None
-        if self.prizes_specs == "algorithm":
-            return self.prizesAlgorithm(buy_in)
-        elif self.prizes_specs == "table":
-            return self.prizesTable(buy_in)
-        else:
-            return None
+        if not self.rank2prize:
+            if self.prizes_specs == "algorithm":
+                self.rank2prize = self.prizesAlgorithm()
+            elif self.prizes_specs == "table":
+                self.rank2prize = self.prizesTable()
+        return self.rank2prize
 
-    def prizesAlgorithm(self, buy_in):
+    def prizesAlgorithm(self):
+        buy_in = self.buy_in
         candidates_count = self.registered
         if candidates_count < 5:
             winners = 1
@@ -434,25 +541,27 @@ class PokerTournament:
             winners = int(candidates_count * 0.1)
 
         prizes = []
-        money_left = buy_in * candidates_count
+        prize_pool = max(self.prize_min, buy_in * candidates_count)
+        money_left = prize_pool
         while winners > 0:
-            if money_left / winners < int(buy_in * 2.5):
+            if money_left / winners < max(1, prize_pool / 100, int(buy_in * 2.5)):
                 prizes.extend([ money_left / winners ] * winners)
                 winners = 0
             else:
                 money_left /= 2
                 winners -= 1
                 prizes.append(money_left)
-        rest = buy_in * candidates_count - sum(prizes)
+        rest = prize_pool - sum(prizes)
         prizes[0] += rest
         return prizes
                 
-    def prizesTable(self, buy_in):
-        for ( max, payouts ) in self.payouts:
-            if self.registered <= max:
+    def prizesTable(self):
+        buy_in = self.buy_in
+        for ( maximum, payouts ) in self.payouts:
+            if self.registered <= maximum:
                 break
 
-        total = self.registered * buy_in
+        total = max(self.prize_min, self.registered * buy_in)
         prizes = map(lambda percent: int(total * percent), payouts)
         #
         # What's left because of rounding errors goes to the tournament winner
